from __future__ import annotations

import typing as t

if t.TYPE_CHECKING:
    import hikari

    from arc.abc.command import CommandProto
    from arc.abc.concurrency_limiting import ConcurrencyLimiterProto
    from arc.abc.limiter import LimiterProto


class ArcError(Exception):
    """Base exception for all Arc errors."""


class AutocompleteError(ArcError):
    """An erro occurred while trying to autocomplete a command."""


class CommandInvokeError(ArcError):
    """An error occurred while trying to invoke a command."""


class ExtensionError(ArcError):
    """A base exception for all extension errors."""


class ExtensionLoadError(ExtensionError):
    """An error occurred while trying to load an extension."""


class ExtensionUnloadError(ExtensionError):
    """An error occurred while trying to unload an extension."""


class InteractionResponseError(ArcError):
    """Base exception for all interaction response errors."""


class HookAbortError(ArcError):
    """Raised when a built-in hook aborts the execution of a command."""


class GuildOnlyError(HookAbortError):
    """Raised when a command is invoked outside of a guild and a
    [`guild_only`][arc.utils.hooks.guild_only] hook is present.
    """


class NotOwnerError(HookAbortError):
    """Raised when a command is invoked by a non-owner and a
    [`owner_only`][arc.utils.hooks.owner_only] hook is present.
    """


class DMOnlyError(HookAbortError):
    """Raised when a command is invoked outside of a DM and a
    [`dm_only`][arc.utils.hooks.dm_only] hook is present.
    """


class InvokerMissingPermissionsError(HookAbortError):
    """Raised when a command is invoked by a user without the
    required permissions set by a [`has_permissions`][arc.utils.hooks.has_permissions] hook.

    Attributes
    ----------
    missing_permissions : hikari.Permissions
        The permissions that the invoker is missing.
    """

    def __init__(self, missing_permissions: hikari.Permissions, *args: t.Any) -> None:
        self.missing_permissions = missing_permissions
        super().__init__(*args)


class BotMissingPermissionsError(HookAbortError):
    """Raised when a command is invoked and the bot is missing the
    required permissions set by a [`bot_has_permissions`][arc.utils.hooks.bot_has_permissions] hook.

    Attributes
    ----------
    missing_permissions : hikari.Permissions
        The permissions that the bot is missing.
    """

    def __init__(self, missing_permissions: hikari.Permissions, *args: t.Any) -> None:
        self.missing_permissions = missing_permissions
        super().__init__(*args)


class NoResponseIssuedError(InteractionResponseError):
    """Raised when no response was issued by a command.
    Interactions must be responded to or deferred within 3 seconds to avoid this error.

    `arc` tries to automatically defer responses when possible, so this error should rarely occur, unless autodefer is disabled.
    """


class ResponseAlreadyIssuedError(InteractionResponseError):
    """Raised when a response was already issued to an interaction.
    Interactions can only be issued one initial response, every other response should be a followup.

    Note that certain actions can only be done in an initial response, such as sending modals or builders.
    """


class UnderCooldownError(ArcError):
    """Raised when a built-in ratelimiter is acquired while it is exhausted, and the
    `wait` parameter is set to `False`.

    Attributes
    ----------
    limiter : arc.abc.limiter.LimiterProto
        The limiter that was rate limited.
    retry_after : float
        The amount of time in seconds until the command is off cooldown.
    """

    def __init__(self, limiter: LimiterProto[t.Any], retry_after: float, *args: t.Any) -> None:
        self.retry_after = retry_after
        self.limiter = limiter
        super().__init__(*args)


class MaxConcurrencyReachedError(ArcError):
    """Raised when a built-in concurrency limiter is acquired while it is exhausted.

    Attributes
    ----------
    limiter : arc.abc.concurrency_limiting.ConcurrencyLimiterProto
        The limiter that was exhausted.
    max_concurrency : int
        The maximum amount of concurrent instances of the command that reached max concurrency.
    """

    def __init__(self, limiter: ConcurrencyLimiterProto[t.Any], max_concurrency: int, *args: t.Any) -> None:
        self.max_concurrency = max_concurrency
        self.limiter = limiter
        super().__init__(*args)


class CommandPublishFailedError(ArcError):
    """Raised when a command could not be published to Discord.

    Attributes
    ----------
    command : arc.abc.command.CommandProto
        The command that failed to publish.
    """

    def __init__(self, command: CommandProto, *args: t.Any) -> None:
        self.command = command
        super().__init__(*args)


class GuildCommandPublishFailedError(ArcError):
    """Raised when a set of commands could not be published to a guild.

    Attributes
    ----------
    guild_id : hikari.Snowflake
        The guild that commands could not be published to.
    """

    def __init__(self, guild_id: hikari.Snowflake, *args: t.Any) -> None:
        self.guild_id = guild_id
        super().__init__(*args)


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



"""A command handler for hikari.

To get started, see the following links:

GitHub:
https://github.com/hypergonial/hikari-arc
Documentation:
https://arc.hypergonial.com
"""

from alluka import Client as Injector
from alluka import inject

from arc import abc, command, ext, utils

from .abc import HookResult, Option, OptionType, with_concurrency_limit, with_hook, with_post_hook
from .client import (
    GatewayClient,
    GatewayClientBase,
    GatewayContext,
    GatewayPlugin,
    RESTClient,
    RESTClientBase,
    RESTContext,
    RESTPlugin,
)
from .command import (
    AttachmentParams,
    BoolParams,
    ChannelParams,
    FloatParams,
    IntParams,
    MentionableParams,
    MessageCommand,
    RoleParams,
    SlashCommand,
    SlashGroup,
    SlashSubCommand,
    SlashSubGroup,
    StrParams,
    UserCommand,
    UserParams,
    message_command,
    slash_command,
    slash_subcommand,
    user_command,
)
from .context import AutocompleteData, AutodeferMode, Context, InteractionResponse
from .errors import (
    ArcError,
    AutocompleteError,
    CommandInvokeError,
    CommandPublishFailedError,
    DMOnlyError,
    ExtensionError,
    ExtensionLoadError,
    ExtensionUnloadError,
    GuildCommandPublishFailedError,
    GuildOnlyError,
    HookAbortError,
    InteractionResponseError,
    InvokerMissingPermissionsError,
    MaxConcurrencyReachedError,
    NoResponseIssuedError,
    NotOwnerError,
    ResponseAlreadyIssuedError,
    UnderCooldownError,
)
from .events import ArcEvent, CommandErrorEvent, StartedEvent, StoppingEvent
from .extension import loader, unloader
from .internal.about import __author__, __author_email__, __license__, __maintainer__, __url__, __version__
from .locale import (
    CommandLocaleRequest,
    CustomLocaleRequest,
    LocaleRequest,
    LocaleRequestType,
    LocaleResponse,
    OptionLocaleRequest,
)
from .plugin import GatewayPluginBase, PluginBase, RESTPluginBase
from .utils import (
    bot_has_permissions,
    channel_concurrency,
    channel_limiter,
    custom_concurrency,
    custom_limiter,
    dm_only,
    global_concurrency,
    global_limiter,
    guild_concurrency,
    guild_limiter,
    guild_only,
    has_permissions,
    member_concurrency,
    member_limiter,
    owner_only,
    user_concurrency,
    user_limiter,
)

__all__ = (
    "__version__",
    "__author__",
    "__author_email__",
    "__license__",
    "__url__",
    "__maintainer__",
    "AutodeferMode",
    "inject",
    "Injector",
    "AutocompleteData",
    "Option",
    "Context",
    "Context",
    "BoolParams",
    "IntParams",
    "StrParams",
    "FloatParams",
    "UserParams",
    "ChannelParams",
    "RoleParams",
    "MentionableParams",
    "AttachmentParams",
    "SlashCommand",
    "SlashGroup",
    "SlashSubCommand",
    "SlashSubGroup",
    "MessageCommand",
    "UserCommand",
    "message_command",
    "user_command",
    "slash_command",
    "slash_subcommand",
    "GatewayClientBase",
    "RESTClientBase",
    "GatewayClient",
    "RESTClient",
    "OptionType",
    "ArcError",
    "AutocompleteError",
    "UnderCooldownError",
    "InvokerMissingPermissionsError",
    "GuildOnlyError",
    "NotOwnerError",
    "DMOnlyError",
    "CommandInvokeError",
    "GuildCommandPublishFailedError",
    "CommandPublishFailedError",
    "MaxConcurrencyReachedError",
    "NoResponseIssuedError",
    "ResponseAlreadyIssuedError",
    "ExtensionError",
    "ExtensionLoadError",
    "ExtensionUnloadError",
    "HookAbortError",
    "InteractionResponseError",
    "PluginBase",
    "RESTPluginBase",
    "GatewayPluginBase",
    "loader",
    "unloader",
    "ArcEvent",
    "CommandErrorEvent",
    "StartedEvent",
    "StoppingEvent",
    "InteractionResponse",
    "GatewayContext",
    "RESTContext",
    "RESTPlugin",
    "GatewayPlugin",
    "HookResult",
    "LocaleRequest",
    "LocaleRequestType",
    "LocaleResponse",
    "CustomLocaleRequest",
    "CommandLocaleRequest",
    "OptionLocaleRequest",
    "abc",
    "utils",
    "command",
    "ext",
    "with_hook",
    "with_post_hook",
    "bot_has_permissions",
    "dm_only",
    "guild_only",
    "has_permissions",
    "owner_only",
    "global_limiter",
    "guild_limiter",
    "with_concurrency_limit",
    "channel_limiter",
    "user_limiter",
    "member_limiter",
    "custom_limiter",
    "global_concurrency",
    "user_concurrency",
    "guild_concurrency",
    "channel_concurrency",
    "member_concurrency",
    "custom_concurrency",
)

# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import abc
import typing as t

from arc.internal.types import ClientT, HasConcurrencyLimiterT

if t.TYPE_CHECKING:
    import typing_extensions as te

    from arc.context.base import Context

KeyT = t.TypeVar("KeyT", contravariant=True)


@t.runtime_checkable
class ConcurrencyLimiterProto(t.Protocol[ClientT]):
    """A protocol for valid concurrency limiters that `arc` can use.

    !!! tip
        An easy (but not necessary) way to ensure you've implemented all methods
        is to inherit from this protocol.
    """

    @property
    @abc.abstractmethod
    def limit(self) -> int:
        """The limit of the concurrency limiter."""

    @abc.abstractmethod
    async def acquire(self, ctx: Context[ClientT], /) -> None:
        """Acquire a concurrency slot for a context.

        This should block until a slot is available.
        """

    @abc.abstractmethod
    def release(self, ctx: Context[ClientT], /) -> None:
        """Release a concurrency slot for a context."""

    @abc.abstractmethod
    def is_exhausted(self, ctx: Context[ClientT], /) -> bool:
        """Return whether the concurrency limiter is exhausted for a context."""


class HasConcurrencyLimiter(abc.ABC, t.Generic[ClientT]):
    """A trait for objects that can have a concurrency limiter."""

    __slots__: t.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def concurrency_limiter(self) -> ConcurrencyLimiterProto[ClientT] | None:
        """The concurrency limiter for this object."""

    @concurrency_limiter.setter
    @abc.abstractmethod
    def concurrency_limiter(self, limiter: ConcurrencyLimiterProto[ClientT] | None) -> None:
        """Set the concurrency limiter for this object."""

    @abc.abstractmethod
    def _resolve_concurrency_limiter(self) -> ConcurrencyLimiterProto[ClientT] | None:
        """Resolve the concurrency limiter for this object."""

    def set_concurrency_limiter(self, limiter: ConcurrencyLimiterProto[ClientT]) -> te.Self:
        """Set the concurrency limiter for this object."""
        self.concurrency_limiter = limiter
        return self


def with_concurrency_limit(
    limiter: ConcurrencyLimiterProto[ClientT],
) -> t.Callable[[HasConcurrencyLimiterT], HasConcurrencyLimiterT]:
    """A decorator that sets a concurrency limiter for an object.

    !!! note
        An object can only have one concurrency limiter set at a time.

    Parameters
    ----------
    limiter : ConcurrencyLimiterProto[ClientT]
        The concurrency limiter to use.

    Returns
    -------
    t.Callable[[HasConcurrencyLimiterT], HasConcurrencyLimiterT]
        The object with the concurrency limiter set.

    Examples
    --------
    ```py
    @client.include
    # Max 5 users can use this command at a time.
    @arc.with_concurrency_limit(arc.user_concurrency(5))
    @arc.slash_command(...)
    async def foo(ctx: arc.GatewayContext) -> None:
        ...
    ```

    See Also
    --------
    - [`guild_concurrency()`][arc.utils.concurrency_limiter.guild_concurrency]
    - [`channel_concurrency()`][arc.utils.concurrency_limiter.channel_concurrency]
    - [`user_concurrency()`][arc.utils.concurrency_limiter.user_concurrency]
    - [`member_concurrency()`][arc.utils.concurrency_limiter.member_concurrency]
    - [`custom_concurrency()`][arc.utils.concurrency_limiter.custom_concurrency]
    """

    def decorator(func: HasConcurrencyLimiterT) -> HasConcurrencyLimiterT:
        func.set_concurrency_limiter(limiter)
        return func

    return decorator


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import sys
import typing as t

import hikari
import pytest
import typing_extensions as te

from arc.abc.option import OptionType
from arc.context import Context
from arc.internal.options import OPTIONTYPE_TO_TYPE


def _get_literals(func: t.Callable[..., t.Any]) -> list[t.Any]:
    """Get all literals found in a function's overload's signatures."""
    literals: list[t.Any] = []
    for overload in te.get_overloads(func):
        hints = t.get_type_hints(overload)
        for hint in hints.values():
            if t.get_origin(hint) is t.Literal:
                # This is a Literal
                args = t.get_args(hint)
                literals.extend(args)
    return literals


def test_optiontype_has_all_hikari_option_types() -> None:
    for option_type in hikari.OptionType:
        try:
            OptionType(int(option_type))
        except Exception as e:
            raise AssertionError(f"Missing {option_type!r} in arc.OptionType enum.") from e


def test_context_get_option_has_all_option_overloads() -> None:
    # Do not test on 3.10-, as it does not support Literal
    if sys.version_info.minor <= 10:
        pytest.skip("Literal is not supported in Python 3.10 or lower")

    literals = _get_literals(Context.get_option)  # type: ignore

    for option_type in OptionType:
        if option_type is OptionType.SUB_COMMAND or option_type is OptionType.SUB_COMMAND_GROUP:
            continue

        assert option_type in literals, f"Missing {option_type!r} in Context.get_option overloads."


def test_options_mapping_contains_all_options() -> None:
    for option_type in OptionType:
        if option_type is OptionType.SUB_COMMAND or option_type is OptionType.SUB_COMMAND_GROUP:
            continue

        assert option_type in OPTIONTYPE_TO_TYPE, f"Missing {option_type!r} in OPTIONTYPE_TO_TYPE mapping."



from __future__ import annotations

import typing as t

import hikari

from arc.abc.plugin import PluginBase
from arc.internal.sigparse import parse_event_signature
from arc.internal.types import EventCallbackT, EventT, GatewayClientT, RESTClientT

if t.TYPE_CHECKING:
    from arc.context.base import AutodeferMode

__all__ = ("RESTPluginBase", "GatewayPluginBase")

P = t.ParamSpec("P")
T = t.TypeVar("T")


class RESTPluginBase(PluginBase[RESTClientT]):
    """The default implementation of a REST plugin.
    To use this with the default [`RESTClient`][arc.client.RESTClient] implementation, see [`RESTPlugin`][arc.client.RESTPlugin].

    Parameters
    ----------
    name : str
        The name of this plugin. This must be unique across all plugins.
    default_enabled_guilds : t.Sequence[hikari.Snowflake] | hikari.UndefinedType
        The default guilds to enable commands in
    autodefer : bool | AutodeferMode
        If True, all commands in this plugin will automatically defer if it is taking longer than 2 seconds to respond.
        This can be overridden on a per-command basis.
    is_dm_enabled : bool | hikari.UndefinedType
        Whether commands in this plugin are enabled in DMs
        This can be overridden on a per-command basis.
    default_permissions : hikari.Permissions | hikari.UndefinedType
        The default permissions for this plugin
        This can be overridden on a per-command basis.
    is_nsfw : bool | hikari.UndefinedType
        Whether this plugin is only usable in NSFW channels

    !!! note
        Parameters left as `hikari.UNDEFINED` will be inherited from the parent client.

    Examples
    --------
    ```py
    plugin = arc.RESTPlugin("MyPlugin")

    @plugin.include
    @arc.slash_command("ping", "Ping the bot.")
    async def ping(ctx: arc.RESTContext) -> None:
        ...

    # Snip

    client.add_plugin(plugin)
    ```
    """

    __slots__: t.Sequence[str] = ()

    @property
    def is_rest(self) -> bool:
        return True


class GatewayPluginBase(PluginBase[GatewayClientT]):
    """The default implementation of a gateway plugin.
    To use this with the default [`GatewayClient`][arc.client.GatewayClient] implementation, see [`GatewayPlugin`][arc.client.GatewayPlugin].

    Parameters
    ----------
    name : str
        The name of this plugin. This must be unique across all plugins.
    default_enabled_guilds : t.Sequence[hikari.Snowflake] | hikari.UndefinedType
        The default guilds to enable commands in
    autodefer : bool | AutodeferMode
        If True, all commands in this plugin will automatically defer if it is taking longer than 2 seconds to respond.
        This can be overridden on a per-command basis.
    is_dm_enabled : bool | hikari.UndefinedType
        Whether commands in this plugin are enabled in DMs
        This can be overridden on a per-command basis.
    default_permissions : hikari.Permissions | hikari.UndefinedType
        The default permissions for this plugin
        This can be overridden on a per-command basis.
    is_nsfw : bool | hikari.UndefinedType
        Whether this plugin is only usable in NSFW channels

    !!! note
        Parameters left as `hikari.UNDEFINED` will be inherited from the parent client.

    Examples
    --------
    ```py
    plugin = arc.GatewayPlugin("MyPlugin")

    @plugin.include
    @arc.slash_command("ping", "Ping the bot.")
    async def ping(ctx: arc.GatewayContext) -> None:
        ...

    # Snip

    client.add_plugin(plugin)
    ```
    """

    __slots__: t.Sequence[str] = ("_listeners",)

    def __init__(
        self,
        name: str,
        *,
        default_enabled_guilds: t.Sequence[hikari.Snowflakeish | hikari.PartialGuild]
        | hikari.UndefinedType = hikari.UNDEFINED,
        autodefer: bool | AutodeferMode | hikari.UndefinedType = hikari.UNDEFINED,
        is_dm_enabled: bool | hikari.UndefinedType = hikari.UNDEFINED,
        default_permissions: hikari.Permissions | hikari.UndefinedType = hikari.UNDEFINED,
        is_nsfw: bool | hikari.UndefinedType = hikari.UNDEFINED,
    ) -> None:
        super().__init__(
            name=name,
            default_enabled_guilds=default_enabled_guilds,
            autodefer=autodefer,
            is_dm_enabled=is_dm_enabled,
            default_permissions=default_permissions,
            is_nsfw=is_nsfw,
        )
        self._listeners: dict[type[hikari.Event], set[EventCallbackT[t.Any]]] = {}

    @property
    def is_rest(self) -> bool:
        return False

    @property
    def listeners(self) -> t.Mapping[type[hikari.Event], t.Collection[EventCallbackT[t.Any]]]:
        return self._listeners

    def subscribe(self, event: type[hikari.Event], callback: EventCallbackT[t.Any]) -> None:
        """Subscribe to an event.

        Parameters
        ----------
        event : type[hikari.Event]
            The event to subscribe to.
        callback : t.Callable[[EventT], t.Awaitable[None]]
            The callback to call when the event is dispatched.
        """
        if event not in self.listeners:
            self._listeners[event] = set()

        self._listeners[event].add(callback)

        if self._client is not None:
            self._client.subscribe(event, callback)

    def unsubscribe(self, event: type[hikari.Event], callback: EventCallbackT[t.Any]) -> None:
        """Unsubscribe from an event.

        Parameters
        ----------
        event : type[hikari.Event]
            The event to unsubscribe from.
        callback : t.Callable[[EventT], t.Awaitable[None]]
            The callback to unsubscribe.
        """
        if event not in self.listeners:
            return

        self._listeners[event].remove(callback)

        if self._client is not None:
            self._client.unsubscribe(event, callback)

    def listen(self, *event_types: type[EventT]) -> t.Callable[[EventCallbackT[EventT]], EventCallbackT[EventT]]:
        """Generate a decorator to subscribe a callback to an event type.

        This is a second-order decorator.

        Parameters
        ----------
        *event_types : type[EventT]
            The event types to subscribe to. If not provided, the event type will be inferred
            instead from the type hints on the function signature.

            `EventT` must be a subclass of `hikari.events.base_events.Event`.

        Returns
        -------
        t.Callable[[EventT], EventT]
            A decorator for a coroutine function that passes it to
            `EventManager.subscribe` before returning the function
            reference.
        """

        def decorator(func: EventCallbackT[EventT]) -> EventCallbackT[EventT]:
            types = event_types or parse_event_signature(func)

            for event_type in types:
                self.subscribe(event_type, func)

            return func

        return decorator

    def _client_include_hook(self, client: GatewayClientT) -> None:
        super()._client_include_hook(client)

        for event, callbacks in self.listeners.items():
            for callback in callbacks:
                client.subscribe(event, callback)

    def _client_remove_hook(self) -> None:
        if self._client is None:
            raise RuntimeError(f"Plugin '{self.name}' is not included in a client.")

        for event, callbacks in self.listeners.items():
            for callback in callbacks:
                self.client.unsubscribe(event, callback)

        super()._client_remove_hook()


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import typing as t

if t.TYPE_CHECKING:
    import hikari

    from arc.abc import Client, Hookable, HookResult, OptionParams
    from arc.abc.concurrency_limiting import HasConcurrencyLimiter
    from arc.client import GatewayClientBase, RESTClientBase
    from arc.command import SlashCommand, SlashGroup
    from arc.context import AutocompleteData, Context
    from arc.locale import CommandLocaleRequest, CustomLocaleRequest, LocaleResponse, OptionLocaleRequest


# Generics
AppT = t.TypeVar("AppT", bound="hikari.RESTAware")
ChoiceT = t.TypeVar("ChoiceT", bound="int | float | str")
ClientT = t.TypeVar("ClientT", bound="Client[t.Any]")
GatewayBotT = t.TypeVar("GatewayBotT", bound="hikari.GatewayBotAware")
RESTBotT = t.TypeVar("RESTBotT", bound="hikari.RESTBotAware")
GatewayClientT = t.TypeVar("GatewayClientT", bound="GatewayClientBase[t.Any]")
RESTClientT = t.TypeVar("RESTClientT", bound="RESTClientBase[t.Any]")
EventT = t.TypeVar("EventT", bound="hikari.Event")
BuilderT = t.TypeVar("BuilderT", bound="hikari.api.SlashCommandBuilder | hikari.api.ContextMenuCommandBuilder")
ParamsT = t.TypeVar("ParamsT", bound="OptionParams[t.Any]")
HookableT = t.TypeVar("HookableT", bound="Hookable[t.Any]")
HasConcurrencyLimiterT = t.TypeVar("HasConcurrencyLimiterT", bound="HasConcurrencyLimiter[t.Any]")

# Type aliases
EventCallbackT: t.TypeAlias = "t.Callable[[EventT], t.Awaitable[None]]"
ErrorHandlerCallbackT: t.TypeAlias = "t.Callable[[Context[ClientT], Exception], t.Awaitable[None]]"
SlashCommandLike: t.TypeAlias = "SlashCommand[ClientT] | SlashGroup[ClientT]"
CommandCallbackT: t.TypeAlias = "t.Callable[t.Concatenate[Context[ClientT], ...], t.Awaitable[None]]"
MessageCommandCallbackT: t.TypeAlias = "t.Callable[[Context[ClientT], hikari.Message], t.Awaitable[None]]"
UserCommandCallbackT: t.TypeAlias = "t.Callable[[Context[ClientT], hikari.User], t.Awaitable[None]]"
AutocompleteCallbackT: t.TypeAlias = "t.Callable[[AutocompleteData[ClientT, ChoiceT]], t.Awaitable[t.Sequence[ChoiceT]]] | t.Callable[[AutocompleteData[ClientT, ChoiceT]], t.Awaitable[t.Sequence[hikari.api.AutocompleteChoiceBuilder]]]"
ResponseBuilderT: t.TypeAlias = (
    "hikari.api.InteractionMessageBuilder | hikari.api.InteractionDeferredBuilder | hikari.api.InteractionModalBuilder"
)
HookT: t.TypeAlias = "t.Callable[[Context[ClientT]], t.Awaitable[HookResult]] | t.Callable[[Context[ClientT]], HookResult] | t.Callable[[Context[ClientT]], None] | t.Callable[[Context[ClientT]], t.Awaitable[None]]"
PostHookT: t.TypeAlias = "t.Callable[[Context[ClientT]], None] | t.Callable[[Context[ClientT]], t.Awaitable[None]]"
LifeCycleHookT: t.TypeAlias = "t.Callable[[ClientT], t.Awaitable[None]]"
CommandLocaleRequestT: t.TypeAlias = "t.Callable[[CommandLocaleRequest], LocaleResponse]"
OptionLocaleRequestT: t.TypeAlias = "t.Callable[[OptionLocaleRequest], LocaleResponse]"
CustomLocaleRequestT: t.TypeAlias = "t.Callable[[CustomLocaleRequest], str]"

# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import asyncio
import contextlib
import typing as t

from arc.abc.concurrency_limiting import ConcurrencyLimiterProto
from arc.context.base import Context
from arc.internal.types import ClientT

__all__ = (
    "ConcurrencyLimiter",
    "CommandConcurrencyLimiter",
    "global_concurrency",
    "guild_concurrency",
    "channel_concurrency",
    "user_concurrency",
    "member_concurrency",
    "custom_concurrency",
)

KeyT = t.TypeVar("KeyT")


# This is cursed
class _BoundedSemaphore(asyncio.BoundedSemaphore):
    @property
    def value(self) -> int:
        return self._value


class _Bucket(t.Generic[KeyT]):
    """Handles the concurrency limiting of a single item. (E.g. a single user or a channel)."""

    __slots__ = ("_key", "_max_concurrent", "_semaphore", "_limiter")

    def __init__(self, key: str, max_concurrent: int, limiter: ConcurrencyLimiter[KeyT]) -> None:
        self._key = key
        self._max_concurrent = max_concurrent
        self._semaphore = _BoundedSemaphore(max_concurrent)
        self._limiter = limiter

    @property
    def key(self) -> str:
        """The key of the bucket."""
        return self._key

    @property
    def remaining(self) -> int:
        """The amount of requests remaining until the bucket is exhausted."""
        return self._semaphore.value

    @property
    def semaphore(self) -> _BoundedSemaphore:
        """The semaphore of the bucket."""
        return self._semaphore

    @classmethod
    def for_limiter(cls, key: str, limiter: ConcurrencyLimiter[KeyT]) -> _Bucket[KeyT]:
        """Create a new bucket for a ConcurrencyLimiter."""
        return cls(key=key, max_concurrent=limiter._capacity, limiter=limiter)

    @property
    def is_exhausted(self) -> bool:
        """Return a boolean determining if the bucket is exhausted."""
        return self._semaphore.locked()

    @property
    def is_stale(self) -> bool:
        """Return a boolean determining if the bucket is stale.
        If a bucket is stale, it is no longer in use and can be purged.
        """
        return self._semaphore.value == self._max_concurrent


class _ConcurrencyLimiterContextManager(t.Generic[KeyT]):
    __slots__ = ("_limiter", "_item")

    def __init__(self, limiter: ConcurrencyLimiter[KeyT], item: KeyT) -> None:
        self._limiter = limiter
        self._item = item

    async def __aenter__(self) -> None:
        await self._limiter.acquire(self._item)

    async def __aexit__(self, exc_type: t.Any, exc_val: t.Any, exc_tb: t.Any) -> None:
        self._limiter.release(self._item)


class ConcurrencyLimiter(t.Generic[KeyT]):
    """A general purpose concurrency limiter.
    Accepts a key extraction function and allocates buckets for each unique key.

    Example:
    ```py
    limiter = arc.utils.ConcurrencyLimiter[int](
        capacity=5,
        get_key_with=lambda x: str(x),
    )

    async with limiter(69):
        print("69 is now being processed")
    ```

    !!! note
        For use with commands, see [`CommandConcurrencyLimiter`][arc.utils.concurrency_limiter.CommandConcurrencyLimiter].

    Parameters
    ----------
    capacity : int
        The maximum amount of concurrently running requests.
    get_key_with : t.Callable[[KeyT], str]
        A callable that returns a key for the concurrency limiter bucket.

    See Also
    --------
    - [`CommandConcurrencyLimiter`][arc.utils.concurrency_limiter.CommandConcurrencyLimiter]
    - [`guild_concurrency()`][arc.utils.concurrency_limiter.guild_concurrency]
    - [`channel_concurrency()`][arc.utils.concurrency_limiter.channel_concurrency]
    - [`user_concurrency()`][arc.utils.concurrency_limiter.user_concurrency]
    - [`member_concurrency()`][arc.utils.concurrency_limiter.member_concurrency]
    - [`custom_concurrency()`][arc.utils.concurrency_limiter.custom_concurrency]
    """

    __slots__: t.Sequence[str] = ("_capacity", "_buckets", "_get_key")

    def __init__(self, capacity: int, *, get_key_with: t.Callable[[KeyT], str]) -> None:
        self._capacity = capacity
        self._buckets: dict[str, _Bucket[KeyT]] = {}
        self._get_key = get_key_with

    def __call__(self, item: KeyT) -> _ConcurrencyLimiterContextManager[KeyT]:
        """Use the limiter as a context manager.

        This will acquire a concurrency slot for an item and release it when the context exits.

        Example:
        ```py
        limiter = arc.utils.ConcurrencyLimiter[str](5, lambda x: x)

        async with limiter("foo"):
            print("foo is now being processed")
        ```

        This is equivalent to:
        ```py
        limiter = arc.utils.ConcurrencyLimiter[str](5, lambda x: x)

        try:
            await limiter.acquire("foo")
            print("foo is now being processed")
        finally:
            await limiter.release("foo")
        ```
        """
        return _ConcurrencyLimiterContextManager(self, item)

    @property
    def limit(self) -> int:
        """The maximum amount of concurrently running requests."""
        return self._capacity

    async def acquire(self, item: KeyT) -> None:
        """Acquire a concurrency slot for an item.

        This will block until a slot is available.
        """
        key = self._get_key(item)
        bucket = self._buckets.setdefault(key, _Bucket.for_limiter(key, self))
        await bucket.semaphore.acquire()

    def release(self, item: KeyT) -> None:
        """Release a concurrency slot for an item."""
        key = self._get_key(item)
        bucket = self._buckets.get(key)
        if bucket is None:
            raise ValueError("Key not found in limiter")

        with contextlib.suppress(ValueError):
            bucket.semaphore.release()

        if bucket.is_stale:
            del self._buckets[key]

    def is_exhausted(self, item: KeyT) -> bool:
        """Return a boolean determining if the limiter is exhausted for an item."""
        key = self._get_key(item)
        bucket = self._buckets.get(key)
        if bucket is None:
            return False
        return bucket.is_exhausted


class CommandConcurrencyLimiter(ConcurrencyLimiter[Context[ClientT]], ConcurrencyLimiterProto[ClientT]):
    """A concurrency limiter specialized for use with `arc` commands.
    This limiter only accepts `arc.context.Context` instances as keys.

    For a generic implementation that works with any key type,
    see [`ConcurrencyLimiter`][arc.utils.concurrency_limiter.ConcurrencyLimiter].

    Parameters
    ----------
    capacity : int
        The maximum amount of concurrently running command instances.
    get_key_with : t.Callable[[Context[ClientT]], str]
        A callable that returns a key for the concurrency limiter bucket.

    See Also
    --------
    - [`guild_concurrency()`][arc.utils.concurrency_limiter.guild_concurrency]
    - [`channel_concurrency()`][arc.utils.concurrency_limiter.channel_concurrency]
    - [`user_concurrency()`][arc.utils.concurrency_limiter.user_concurrency]
    - [`member_concurrency()`][arc.utils.concurrency_limiter.member_concurrency]
    - [`custom_concurrency()`][arc.utils.concurrency_limiter.custom_concurrency]
    """

    __slots__ = ()

    def __call__(self, item: Context[ClientT]) -> _ConcurrencyLimiterContextManager[Context[ClientT]]:
        """Use the limiter as a context manager.

        This will acquire a concurrency slot for an item and release it when the context exits.

        Example:
        ```py
        limiter = arc.utils.CommandConcurrencyLimiter(5, lambda ctx: str(ctx.guild_id)))

        async with limiter(ctx):
            print("ctx is now being processed")
        ```

        This is equivalent to:
        ```py
        limiter = arc.utils.CommandConcurrencyLimiter(5, lambda ctx: str(ctx.guild_id)))

        try:
            await limiter.acquire(ctx)
            print("ctx is now being processed")
        finally:
            await limiter.release(ctx)
        ```
        """
        return super().__call__(item)


def global_concurrency() -> CommandConcurrencyLimiter[t.Any]:
    """Limit a command to a certain amount of concurrent instances globally.

    Returns
    -------
    CommandConcurrencyLimiter[t.Any]
        A concurrency limiter for use with a command.

    Examples
    --------
    ```py
    @arc.with_concurrency_limit(arc.global_concurrency())
    ```
    """
    return CommandConcurrencyLimiter(1, get_key_with=lambda _: "amongus")


def guild_concurrency(limit: int) -> CommandConcurrencyLimiter[t.Any]:
    """Limit a command to a certain amount of concurrent instances per guild.

    Parameters
    ----------
    limit : int
        The maximum amount of concurrently running command instances.

    Returns
    -------
    CommandConcurrencyLimiter[t.Any]
        A concurrency limiter for use with a command.

    Examples
    --------
    ```py
    @arc.with_concurrency_limit(arc.guild_concurrency(1))
    ```
    """
    return CommandConcurrencyLimiter(limit, get_key_with=lambda ctx: str(ctx.guild_id))


def channel_concurrency(limit: int) -> CommandConcurrencyLimiter[t.Any]:
    """Limit a command to a certain amount of concurrent instances per channel.

    Parameters
    ----------
    limit : int
        The maximum amount of concurrently running command instances.

    Returns
    -------
    CommandConcurrencyLimiter[t.Any]
        A concurrency limiter for use with a command.

    Examples
    --------
    ```py
    @arc.with_concurrency_limit(arc.channel_concurrency(1))
    ```
    """
    return CommandConcurrencyLimiter(limit, get_key_with=lambda ctx: str(ctx.channel_id))


def user_concurrency(limit: int) -> CommandConcurrencyLimiter[t.Any]:
    """Limit a command to a certain amount of concurrent instances per user.

    Parameters
    ----------
    limit : int
        The maximum amount of concurrently running command instances.

    Returns
    -------
    CommandConcurrencyLimiter[t.Any]
        A concurrency limiter for use with a command.

    Examples
    --------
    ```py
    @arc.with_concurrency_limit(arc.user_concurrency(1))
    ```
    """
    return CommandConcurrencyLimiter(limit, get_key_with=lambda ctx: str(ctx.author.id))


def member_concurrency(limit: int) -> CommandConcurrencyLimiter[t.Any]:
    """Limit a command to a certain amount of concurrent instances per member.

    Parameters
    ----------
    limit : int
        The maximum amount of concurrently running command instances.

    Returns
    -------
    CommandConcurrencyLimiter[t.Any]
        A concurrency limiter for use with a command.

    Examples
    --------
    ```py
    @arc.with_concurrency_limit(arc.member_concurrency(1))
    ```
    """
    return CommandConcurrencyLimiter(limit, get_key_with=lambda ctx: f"{ctx.guild_id}:{ctx.author.id}")


def custom_concurrency(
    limit: int, get_key_with: t.Callable[[Context[ClientT]], str]
) -> CommandConcurrencyLimiter[t.Any]:
    """Limit a command to a certain amount of concurrent instances per custom key.

    Parameters
    ----------
    limit : int
        The maximum amount of concurrently running command instances.
    get_key_with : t.Callable[[Context[ClientT]], str]
        A callable that returns a key for the concurrency limiter bucket.

    Returns
    -------
    CommandConcurrencyLimiter[t.Any]
        A concurrency limiter for use with a command.

    Examples
    --------
    ```py
    # This is identical to 'arc.guild_concurrency(1)'
    @arc.with_concurrency_limit(arc.custom_concurrency(1, lambda ctx: str(ctx.guild_id)))
    ```
    """
    return CommandConcurrencyLimiter(limit, get_key_with=get_key_with)


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import typing as t

import attr
import hikari

from arc.abc.command import CallableCommandBase
from arc.context import AutodeferMode, Context
from arc.errors import CommandInvokeError
from arc.internal.types import ClientT, ResponseBuilderT

if t.TYPE_CHECKING:
    import asyncio

    from arc.abc.command import CallableCommandProto
    from arc.internal.types import UserCommandCallbackT

__all__ = ("UserCommand", "user_command")


@attr.define(slots=True, kw_only=True)
class UserCommand(CallableCommandBase[ClientT, hikari.api.ContextMenuCommandBuilder]):
    """A context menu command that is invoked by right-clicking a user."""

    @property
    def command_type(self) -> hikari.CommandType:
        return hikari.CommandType.USER

    @property
    def qualified_name(self) -> t.Sequence[str]:
        return (self.name,)

    def _get_context(
        self, interaction: hikari.CommandInteraction, command: CallableCommandProto[ClientT]
    ) -> Context[ClientT]:
        assert self.client is not None

        return Context(self.client, command, interaction)

    def _build(self) -> hikari.api.ContextMenuCommandBuilder:
        return hikari.impl.ContextMenuCommandBuilder(
            name=self.name,
            type=self.command_type,
            default_member_permissions=self.default_permissions,
            is_dm_enabled=self.is_dm_enabled,
            is_nsfw=self.is_nsfw,
            name_localizations=self.name_localizations,  # pyright: ignore reportGeneralTypeIssues
        )

    async def invoke(
        self, interaction: hikari.CommandInteraction, *args: t.Any, **kwargs: t.Any
    ) -> asyncio.Future[ResponseBuilderT] | None:
        if interaction.command_type is not hikari.CommandType.USER:
            raise CommandInvokeError("Cannot invoke a user command with a non-user command interaction.")

        assert interaction.resolved is not None and interaction.target_id is not None

        user = (
            interaction.resolved.members.get(interaction.target_id) or interaction.resolved.users[interaction.target_id]
        )

        return await super().invoke(interaction, user, *args, **kwargs)


def user_command(
    name: str,
    *,
    guilds: t.Sequence[hikari.PartialGuild | hikari.Snowflakeish] | hikari.UndefinedType = hikari.UNDEFINED,
    is_dm_enabled: bool | hikari.UndefinedType = hikari.UNDEFINED,
    is_nsfw: bool | hikari.UndefinedType = hikari.UNDEFINED,
    autodefer: bool | AutodeferMode | hikari.UndefinedType = hikari.UNDEFINED,
    default_permissions: hikari.Permissions | hikari.UndefinedType = hikari.UNDEFINED,
    name_localizations: t.Mapping[hikari.Locale, str] | None = None,
) -> t.Callable[[UserCommandCallbackT[ClientT]], UserCommand[ClientT]]:
    """A decorator that creates a context-menu command on a user.

    Parameters
    ----------
    name : str
        The name of the command.
    guilds : t.Sequence[hikari.PartialGuild | hikari.Snowflakeish] | hikari.UndefinedType
        The guilds this command should be enabled in, if left as undefined, the command is global
    is_dm_enabled : bool | hikari.UndefinedType
        Whether this command is enabled in DMs.
    is_nsfw : bool | hikari.UndefinedType
        Whether this command is NSFW.
    autodefer : bool | AutodeferMode | hikari.UndefinedType
        If True, this command will be automatically deferred if it takes longer than 2 seconds to respond
    default_permissions : hikari.Permissions | hikari.UndefinedType
        The default permissions for this command.
        Keep in mind that guild administrators can change this, it should only be used to provide safe defaults.
    name_localizations : t.Mapping[hikari.Locale, str] | None
        The localizations for this command's name.

    !!! note
        Parameters left as `hikari.UNDEFINED` will be inherited from the parent plugin or client.

    Examples
    --------
    ```py
    @client.include
    @arc.user_command(name="Say Hi", description="Say hi!")
    async def hi_user(ctx: arc.GatewayContext, user: hikari.User) -> None:
        await ctx.respond(f"Hey {user.mention}!")
    ```
    """

    def decorator(callback: UserCommandCallbackT[ClientT]) -> UserCommand[ClientT]:
        guild_ids = tuple(hikari.Snowflake(i) for i in guilds) if guilds is not hikari.UNDEFINED else hikari.UNDEFINED

        return UserCommand(
            callback=callback,
            name=name,
            autodefer=AutodeferMode(autodefer) if isinstance(autodefer, bool) else autodefer,
            guilds=guild_ids,
            is_dm_enabled=is_dm_enabled,
            is_nsfw=is_nsfw,
            default_permissions=default_permissions,
            name_localizations=name_localizations or {},
        )

    return decorator


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import hikari

import arc

bot = hikari.GatewayBot("...", banner=None)
client = arc.GatewayClient(bot)


@client.include
@arc.user_command(name="Say Hi")
async def ping_user(ctx: arc.GatewayContext, user: hikari.User) -> None:
    await ctx.respond(f"Hi {user}!")


@client.include()
@arc.message_command(name="Say Hi")
async def ping_message(ctx: arc.GatewayContext, message: hikari.Message) -> None:
    await ctx.respond(f"Hi {message.author}!")


def test_user_command():
    assert len(client._user_commands) == 1
    command = client._user_commands["Say Hi"]

    assert command.name == "Say Hi"
    assert command.command_type is hikari.CommandType.USER
    assert command.qualified_name == ("Say Hi",)
    assert command.default_permissions is hikari.UNDEFINED
    assert command.is_dm_enabled is True
    assert command.is_nsfw is False
    assert command.name_localizations == {}


def test_message_command():
    assert len(client._message_commands) == 1
    command = client._message_commands["Say Hi"]

    assert command.name == "Say Hi"
    assert command.command_type is hikari.CommandType.MESSAGE
    assert command.qualified_name == ("Say Hi",)
    assert command.default_permissions is hikari.UNDEFINED
    assert command.is_dm_enabled is True
    assert command.is_nsfw is False
    assert command.name_localizations == {}



from __future__ import annotations

import abc
import enum
import typing as t
from typing import Any

import attr
import hikari

from arc.internal.types import AutocompleteCallbackT, ChoiceT, ClientT, ParamsT
from arc.locale import OptionLocaleRequest

if t.TYPE_CHECKING:
    import typing_extensions as te

    from arc.abc.client import Client
    from arc.abc.command import CommandProto

__all__ = (
    "Option",
    "OptionParams",
    "OptionWithChoices",
    "OptionWithChoicesParams",
    "OptionBase",
    "CommandOptionBase",
    "OptionType",
)

T = t.TypeVar("T")

Option = t.Annotated
"""Alias for typing.Annotated.

Examples
--------
```py
arc.Option[type, arc.TypeParams(...)]
```

So for example, to create an `int` option, you would do:

```py
arc.Option[int, arc.IntParams(...)]
```
"""


class OptionType(enum.IntEnum):
    """The type of a command option.

    This is practically identical to `hikari.OptionType` at the moment.
    It may however be used in the future to define custom option types.
    """

    SUB_COMMAND = 1
    """Denotes a command option where the value will be a sub command."""

    SUB_COMMAND_GROUP = 2
    """Denotes a command option where the value will be a sub command group."""

    STRING = 3
    """Denotes a command option where the value will be a string."""

    INTEGER = 4
    """Denotes a command option where the value will be a int.

    This is range limited between -2^53 and 2^53.
    """

    BOOLEAN = 5
    """Denotes a command option where the value will be a bool."""

    USER = 6
    """Denotes a command option where the value will be resolved to a user."""

    CHANNEL = 7
    """Denotes a command option where the value will be resolved to a channel."""

    ROLE = 8
    """Denotes a command option where the value will be resolved to a role."""

    MENTIONABLE = 9
    """Denotes a command option where the value will be a snowflake ID."""

    FLOAT = 10
    """Denotes a command option where the value will be a float.

    This is range limited between -2^53 and 2^53.
    """

    ATTACHMENT = 11
    """Denotes a command option where the value will be an attachment."""

    @classmethod
    def from_hikari(cls, option_type: hikari.OptionType) -> OptionType:
        """Convert a hikari.OptionType to an OptionType."""
        return cls(option_type.value)

    def to_hikari(self) -> hikari.OptionType:
        """Convert an OptionType to a hikari.OptionType."""
        # TODO: Map custom option types to their respective hikari.OptionType
        return hikari.OptionType(self.value)

    # TODO: When adding custom convertible option types, add them with an offset of 1000 or so


class OptionParams(t.Generic[T]):
    """The base class for all option parameters objects.

    Parameters
    ----------
    name : str
        The name of the option
    description : str
        The description of the option
    name_localizations : Mapping[hikari.Locale, str]
        The name of the option in different locales
    description_localizations : Mapping[hikari.Locale, str]
        The description of the option in different locales
    """

    __slots__: t.Sequence[str] = ("_name", "_description", "_name_localizations", "_description_localizations")

    def __init__(
        self,
        description: str = "No description provided.",
        *,
        name: str | None = None,
        name_localizations: t.Mapping[hikari.Locale, str] | None = None,
        description_localizations: t.Mapping[hikari.Locale, str] | None = None,
    ) -> None:
        self._description = description
        self._name = name
        self._name_localizations = name_localizations or {}
        self._description_localizations = description_localizations or {}

    @property
    def description(self) -> str:
        """The description of the option."""
        return self._description

    @property
    def name(self) -> str | None:
        """The name of the option."""
        return self._name

    @property
    def name_localizations(self) -> t.Mapping[hikari.Locale, str]:
        """The name of the option in different locales."""
        return self._name_localizations

    @property
    def description_localizations(self) -> t.Mapping[hikari.Locale, str]:
        """The description of the option in different locales."""
        return self._description_localizations


class OptionWithChoicesParams(OptionParams[ChoiceT], t.Generic[ChoiceT, ClientT]):
    """The parameters for an option that can have choices or be autocompleted.

    Parameters
    ----------
    name : str
        The name of the option
    description : str
        The description of the option
    name_localizations : Mapping[hikari.Locale, str]
        The name of the option in different locales
    description_localizations : Mapping[hikari.Locale, str]
        The description of the option in different locales
    choices : t.Sequence[ChoiceT | hikari.CommandChoice] | t.Mapping[str, ChoiceT] | None
        The choices for the option. If provided, these will be the only valid values for the option.
    autocomplete_with : AutocompleteCallbackT[ClientT, ChoiceT] | None
        The callback for autocompleting the option.

    !!! warning
        You cannot provide both `choices` and `autocomplete_with` at the same time.
    """

    __slots__: t.Sequence[str] = ("_choices", "_autocomplete_with")

    def __init__(
        self,
        description: str = "No description provided.",
        *,
        name: str | None = None,
        name_localizations: t.Mapping[hikari.Locale, str] | None = None,
        description_localizations: t.Mapping[hikari.Locale, str] | None = None,
        choices: t.Sequence[ChoiceT | hikari.CommandChoice] | t.Mapping[str, ChoiceT] | None = None,
        autocomplete_with: AutocompleteCallbackT[ClientT, ChoiceT] | None = None,
    ) -> None:
        super().__init__(
            description=description,
            name=name,
            name_localizations=name_localizations,
            description_localizations=description_localizations,
        )
        self._choices = choices
        self._autocomplete_with = autocomplete_with

    @property
    def choices(self) -> t.Sequence[ChoiceT | hikari.CommandChoice] | t.Mapping[str, ChoiceT] | None:
        """The choices for the option. If provided, these will be the only valid values for the option."""
        return self._choices

    @property
    def autocomplete_with(self) -> AutocompleteCallbackT[ClientT, ChoiceT] | None:
        """The callback for autocompleting the option."""
        return self._autocomplete_with


@attr.define(slots=True, kw_only=True)
class OptionBase(abc.ABC, t.Generic[T]):
    """An abstract base class for all slash options and subcommands."""

    name: str
    """The name of the option."""
    description: str
    """The description of the option."""
    name_localizations: t.Mapping[hikari.Locale, str] = attr.field(factory=dict)
    """The name of the option in different locales."""
    description_localizations: t.Mapping[hikari.Locale, str] = attr.field(factory=dict)
    """The description of the option in different locales."""

    @property
    @abc.abstractmethod
    def option_type(self) -> OptionType:
        """The type of the option. Used to register the command."""

    def _to_dict(self) -> dict[str, t.Any]:
        """Convert the option to a dictionary of kwargs that can be passed to hikari.CommandOption."""
        return {
            "type": self.option_type.to_hikari(),
            "name": self.name,
            "description": self.description,
            "autocomplete": False,
            "name_localizations": self.name_localizations,
            "description_localizations": self.description_localizations,
        }

    def to_command_option(self) -> hikari.CommandOption:
        """Convert this option to a hikari.CommandOption."""
        return hikari.CommandOption(**self._to_dict())

    def _request_option_locale(self, client: Client[t.Any], command: CommandProto) -> None:
        """Request the option's name and description in different locales."""
        if self.name_localizations or self.description_localizations:
            return

        if not client._provided_locales or not client._option_locale_provider:
            return

        name_locales: dict[hikari.Locale, str] = {}
        desc_locales: dict[hikari.Locale, str] = {}

        for locale in client._provided_locales:
            request = OptionLocaleRequest(command, locale, self.name, self.description, self)
            resp = client._option_locale_provider(request)

            if resp.name is not None and resp.description is not None:
                name_locales[locale] = resp.name
                desc_locales[locale] = resp.description

        self.name_localizations = name_locales
        self.description_localizations = desc_locales


@attr.define(slots=True, kw_only=True)
class CommandOptionBase(OptionBase[T], t.Generic[T, ClientT, ParamsT]):
    """An abstract base class for all slash command options. This does not include subcommands."""

    is_required: bool = True
    """Whether the option is required."""

    @classmethod
    @abc.abstractmethod
    def _from_params(cls, *, name: str, is_required: bool, params: ParamsT, **kwargs: t.Any) -> te.Self:
        """Construct a new Option instance from the given parameters object.

        Parameters
        ----------
        name : str
            The name of the option
        is_required : bool
            Whether the option is required
        params : ParamsT
            The parameters for the option
        kwargs : dict[str, Any]
            Any additional keyword arguments to pass to the constructor
        """

    def _to_dict(self) -> dict[str, Any]:
        return {**super()._to_dict(), "is_required": self.is_required}


@attr.define(slots=True, kw_only=True)
class OptionWithChoices(CommandOptionBase[ChoiceT, ClientT, ParamsT]):
    """An option that can have choices or be autocompleted."""

    choices: t.Sequence[ChoiceT | hikari.CommandChoice] | t.Mapping[str, ChoiceT] | None = None
    """The choices for the option."""

    autocomplete_with: AutocompleteCallbackT[ClientT, ChoiceT] | None = None
    """The callback for autocompleting the option."""

    def _choices_to_command_choices(self) -> t.Sequence[hikari.CommandChoice] | None:
        if self.choices is None:
            return None

        if isinstance(self.choices, t.Mapping):
            return [hikari.CommandChoice(name=str(name), value=value) for name, value in self.choices.items()]

        return [
            hikari.CommandChoice(name=str(choice), value=choice)
            if not isinstance(choice, hikari.CommandChoice)
            else choice
            for choice in self.choices
        ]

    def _to_dict(self) -> dict[str, t.Any]:
        return {
            **super()._to_dict(),
            "choices": self._choices_to_command_choices(),
            "autocomplete": self.autocomplete_with is not None,
        }


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import inspect
import sys
import types
import typing as t

import hikari

from arc.abc.option import OptionParams
from arc.command.option import (
    AttachmentOption,
    AttachmentParams,
    BoolOption,
    BoolParams,
    ChannelOption,
    ChannelParams,
    FloatOption,
    FloatParams,
    IntOption,
    IntParams,
    MentionableOption,
    MentionableParams,
    RoleOption,
    RoleParams,
    StrOption,
    StrParams,
    UserOption,
    UserParams,
)

if t.TYPE_CHECKING:
    from arc.abc.option import CommandOptionBase
    from arc.context import Context
    from arc.internal.types import ClientT, EventT

# pyright: reportUnnecessaryTypeIgnoreComment=false

__all__ = ("parse_command_signature",)

# Potential pyright bug? This wasn't reported in 1.1.345
#
# This doesn't include some special cases, for the complete resolution logic see: _get_option_type()
TYPE_TO_OPTION_MAPPING: dict[type[t.Any], type[CommandOptionBase[t.Any, t.Any, t.Any]]] = {
    bool: BoolOption,
    int: IntOption,
    str: StrOption,
    float: FloatOption,
    hikari.Role: RoleOption,
    hikari.Attachment: AttachmentOption,
    hikari.User: UserOption,
}

OPT_TO_PARAMS_MAPPING: dict[type[CommandOptionBase[t.Any, t.Any, t.Any]], type[t.Any]] = {
    BoolOption: BoolParams,
    IntOption: IntParams,
    StrOption: StrParams,
    FloatOption: FloatParams,
    UserOption: UserParams,
    ChannelOption: ChannelParams,
    MentionableOption: MentionableParams,
    RoleOption: RoleParams,
    AttachmentOption: AttachmentParams,
}

BASE_CHANNEL_TYPE_MAP: dict[type[hikari.PartialChannel], hikari.ChannelType] = {
    hikari.GuildTextChannel: hikari.ChannelType.GUILD_TEXT,
    hikari.DMChannel: hikari.ChannelType.DM,
    hikari.GuildVoiceChannel: hikari.ChannelType.GUILD_VOICE,
    hikari.GroupDMChannel: hikari.ChannelType.GROUP_DM,
    hikari.GuildCategory: hikari.ChannelType.GUILD_CATEGORY,
    hikari.GuildNewsChannel: hikari.ChannelType.GUILD_NEWS,
    hikari.GuildNewsThread: hikari.ChannelType.GUILD_NEWS_THREAD,
    hikari.GuildPublicThread: hikari.ChannelType.GUILD_PUBLIC_THREAD,
    hikari.GuildPrivateThread: hikari.ChannelType.GUILD_PRIVATE_THREAD,
    hikari.GuildStageChannel: hikari.ChannelType.GUILD_STAGE,
    hikari.GuildForumChannel: hikari.ChannelType.GUILD_FORUM,
}


def _get_channel_type(channel: type[hikari.PartialChannel]) -> set[hikari.ChannelType]:
    """Get channel types from a channel."""
    if channel in (hikari.PartialChannel, hikari.InteractionChannel):
        return set()

    types: set[hikari.ChannelType] = set()
    for k, v in BASE_CHANNEL_TYPE_MAP.items():
        if issubclass(k, channel):
            types.add(v)

    return types


def _get_all_channel_types() -> dict[type[hikari.PartialChannel], set[hikari.ChannelType]]:
    """Get all channels and their corresponding channel types."""
    mapping: dict[type[hikari.PartialChannel], set[hikari.ChannelType]] = {}

    for _, attribute in inspect.getmembers(
        hikari, lambda a: isinstance(a, type) and issubclass(a, hikari.PartialChannel)
    ):
        mapping[attribute] = _get_channel_type(attribute)

    return mapping


# Python macros when
CHANNEL_TYPES_MAPPING = _get_all_channel_types()


def _get_option_type(hint: t.Any) -> type[CommandOptionBase[t.Any, t.Any, t.Any]] | None:
    """Get the option type from a type hint."""
    if _is_mentionable_union(hint):
        return MentionableOption  # pyright: ignore reportGeneralTypeIssues

    elif _is_union(hint):
        hints = [arg for arg in t.get_args(hint) if arg is not type(None)]
        first = _get_option_type(hints[0])
        # Check if it is a uniform union recursively
        if all(_get_option_type(arg) is first for arg in hints):
            return first

    elif hint in CHANNEL_TYPES_MAPPING:
        return ChannelOption  # pyright: ignore reportGeneralTypeIssues

    else:
        return TYPE_TO_OPTION_MAPPING.get(hint)


def _is_param(meta: t.Any) -> bool:
    """Return True if the metadata is a command option parameter object."""
    return isinstance(meta, OptionParams)


def _is_union(hint: t.Any) -> bool:
    """Return True if the type hint is a typing.Union. or Python 3.10's types.UnionType."""
    return t.get_origin(hint) is t.Union or t.get_origin(hint) is types.UnionType


def _is_optional_union(hint: t.Any) -> bool:
    """Return True if the type hint is a typing.Union[T, None], also known as typing.Optional[T]."""
    return t.get_origin(hint) is t.Union and len(t.get_args(hint)) == 2 and type(None) in t.get_args(hint)


def _extract_optional_type(hint: t.Any) -> type[t.Any]:
    """Convert typing.Optional[T] to T."""
    return next(arg for arg in t.get_args(hint) if arg is not type(None))


def _is_mentionable_union(hint: t.Any) -> bool:
    """Check if a type hint is a union that represents a MentionableOption.

    Parameters
    ----------
    hint : t.Any
        The type hint to check

    Returns
    -------
    bool
        Whether the type hint is a union that represents a Mentionable option
    """
    if not _is_union(hint):
        return False

    return {arg for arg in t.get_args(hint) if arg is not type(None)} == {hikari.User, hikari.Role}


def _channels_to_channel_types(channels: t.Iterable[type[hikari.PartialChannel]]) -> list[hikari.ChannelType]:
    """Turn a list of channels into a list of channel types.

    Parameters
    ----------
    channels : Iterable[hikari.PartialChannel]
        The channels to parse

    Returns
    -------
    list[hikari.ChannelType]
        The list of channel types
    """
    channel_types: set[hikari.ChannelType] = set()

    for channel in channels:
        types = CHANNEL_TYPES_MAPPING.get(channel)

        if types is None:
            raise TypeError(
                f"Unsupported channel type '{channel.__name__}'\nSupported types: {tuple(CHANNEL_TYPES_MAPPING)}"
            )

        channel_types.update(types)

    return list(channel_types)


def _parse_channel_union_type_hint(hint: t.Any) -> list[hikari.ChannelType]:
    """Turn a union of channel types into a list of channel types.

    E.g. `hikari.GuildTextChannel | hikari.DMChannel` becomes `[hikari.ChannelType.GUILD_TEXT, hikari.ChannelType.DM]`

    Parameters
    ----------
    hint : t.Any
        The type hint to parse

    Returns
    -------
    list[hikari.ChannelType]
        The list of channel types

    Raises
    ------
    TypeError
        The type hint was not a union of channel types
    """
    if not _is_union(hint):
        raise TypeError(f"Expected type hint to be 'Union', got '{hint!r}'")

    args = t.get_args(hint)

    if not all((issubclass(arg, hikari.PartialChannel)) or arg is type(None) for arg in args):
        raise TypeError(f"Union of channels is not uniform: '{hint!r}'")

    return _channels_to_channel_types(arg for arg in args if arg is not type(None))


def parse_command_signature(  # noqa: C901
    func: t.Callable[t.Concatenate[Context[ClientT], ...], t.Awaitable[None]],
) -> dict[str, CommandOptionBase[t.Any, t.Any, t.Any]]:
    """Parse a command callback function's signature and return a list of options.

    Parameters
    ----------
    func : Callable[Concatenate[Context, P], None]
        The callback function to parse

    Returns
    -------
    dict[str, Option]
        A mapping of the keyword-argument name to the parsed option

    Raises
    ------
    TypeError
        One of the type hints was not of type 'Annotated'
    TypeError
        The type hint had more than 1 metadata item
    TypeError
        Unsupported option type
    TypeError
        The params type does not match the argument type
    TypeError
        One of the union types was not supported
    """
    hints = t.get_type_hints(func, include_extras=True)
    parameters = inspect.signature(func).parameters

    options: dict[str, CommandOptionBase[t.Any, t.Any, t.Any]] = {}
    # Remove the return type
    hints.pop("return", None)

    for arg_name, hint in hints.items():
        hint: t.Any

        # Ignore non-annotated type hints
        if t.get_origin(hint) is not t.Annotated:
            # Python 3.10 has this funny behaviour where if you default a parameter to `None`,
            # it will automatically wrap it in `typing.Optional` because fuck you.
            # So we will just get the value out of the Optional if it is in one
            if tuple(sys.version_info)[:2] == (3, 10) and _is_optional_union(hint):
                hint = _extract_optional_type(hint)

                if t.get_origin(hint) is not t.Annotated:
                    continue
            else:
                continue

        if len(hint.__metadata__) != 1:
            continue

        params = hint.__metadata__[0]
        type_ = t.get_args(hint)[0]
        union = None
        is_optional = parameters[arg_name].default is not inspect.Parameter.empty

        # This may be an alluka.Injected or other type of annotation
        if not _is_param(params):
            continue

        # If it's a union, update is_optional
        if _is_union(type_):
            union = type_
            is_optional = is_optional or type(None) in t.get_args(union)

        opt_type = _get_option_type(type_)

        # If the opt_type is None, it failed to resolve
        if opt_type is None:
            raise TypeError(f"Unsupported option type: '{type_!r}'")

        # Verify the params type matches the option type
        if not isinstance(params, OPT_TO_PARAMS_MAPPING[opt_type]):
            raise TypeError(
                f"Expected params object to be of type '{OPT_TO_PARAMS_MAPPING[opt_type].__name__}', got '{type(params).__name__}'"
            )

        # If it's a union of channel types, we need to parse all channel types
        if union is not None and any(arg in CHANNEL_TYPES_MAPPING for arg in t.get_args(union)):
            channel_types = _parse_channel_union_type_hint(union)
            options[arg_name] = ChannelOption._from_params(
                name=params.name or arg_name, is_required=not is_optional, params=params, channel_types=channel_types
            )
            continue

        # If it's a single channel type, just pass the channel type
        elif type_ in CHANNEL_TYPES_MAPPING:
            options[arg_name] = ChannelOption._from_params(
                name=params.name or arg_name,
                is_required=not is_optional,
                params=params,
                channel_types=_channels_to_channel_types([type_]),
            )
            continue

        # Otherwise just build the option
        options[arg_name] = opt_type._from_params(
            name=params.name or arg_name, is_required=not is_optional, params=params
        )

    return options


def parse_event_signature(func: t.Callable[[EventT], t.Awaitable[None]]) -> list[type[EventT]]:
    """Parse an event callback function's signature and return the event type, ignore other type hints."""
    hints = t.get_type_hints(func)

    # Remove the return type
    hints.pop("return", None)

    first = next(iter(hints.values()))

    if _is_union(first):
        events = [arg for arg in t.get_args(first) if issubclass(arg, hikari.Event)]
        if not events:
            raise TypeError("Expected event callback to have first argument that inherits from 'hikari.Event'")
        return events  # pyright: ignore reportGeneralTypeIssues

    elif issubclass(first, hikari.Event):
        return [first]  # pyright: ignore reportGeneralTypeIssues

    raise TypeError(
        f"Expected event callback to have first argument that inherits from 'hikari.Event', got '{first!r}'"
    )


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import hikari

import arc

bot = hikari.GatewayBot("...", banner=None)
client = arc.GatewayClient(bot)

group = client.include_slash_group(
    "my_group", "My group description", default_permissions=hikari.Permissions.ADMINISTRATOR
)

subgroup = group.include_subgroup("my_subgroup", "My subgroup description")


@client.include
@arc.message_command("Message Command")
async def message_cmd(ctx: arc.GatewayContext, message: hikari.Message) -> None:
    pass


@group.include
@arc.slash_subcommand("test_subcommand", "My subcommand description")
async def my_subcommand(
    ctx: arc.GatewayContext,
    a: arc.Option[int, arc.IntParams(description="foo", min=10)],
    b: arc.Option[str, arc.StrParams(description="bar", min_length=100)],
) -> None:
    pass


@subgroup.include()
@arc.slash_subcommand("test_subsubcommand", "My subsubcommand description")
async def my_subsubcommand(
    ctx: arc.GatewayContext,
    a: arc.Option[int, arc.IntParams(description="foo", min=10)],
    b: arc.Option[str, arc.StrParams(description="bar", min_length=100)],
) -> None:
    pass


def test_walk_commands() -> None:
    cmds = list(client.walk_commands(hikari.CommandType.SLASH, callable_only=False))

    assert len(cmds) == 4

    assert my_subcommand in cmds
    assert my_subsubcommand in cmds
    assert group in cmds
    assert subgroup in cmds

    cmds = list(client.walk_commands(hikari.CommandType.SLASH, callable_only=True))

    assert len(cmds) == 2

    assert my_subcommand in cmds
    assert my_subsubcommand in cmds

    cmds = list(client.walk_commands(hikari.CommandType.MESSAGE))

    assert len(cmds) == 1

    assert message_cmd in cmds



from .basic import bot_has_permissions, dm_only, guild_only, has_permissions, owner_only
from .limiters import (
    LimiterHook,
    channel_limiter,
    custom_limiter,
    global_limiter,
    guild_limiter,
    member_limiter,
    user_limiter,
)

__all__ = (
    "guild_only",
    "owner_only",
    "dm_only",
    "has_permissions",
    "bot_has_permissions",
    "global_limiter",
    "guild_limiter",
    "channel_limiter",
    "user_limiter",
    "member_limiter",
    "custom_limiter",
    "LimiterHook",
)



from __future__ import annotations

import typing as t

import attr

from arc.abc.option import OptionType, OptionWithChoices, OptionWithChoicesParams
from arc.internal.types import ClientT

if t.TYPE_CHECKING:
    import hikari
    import typing_extensions as te

    from arc.internal.types import AutocompleteCallbackT


__all__ = ("StrOption", "StrParams")


@t.final
class StrParams(OptionWithChoicesParams[str, ClientT]):
    """The parameters for a string option.

    Parameters
    ----------
    description : str
        The description of the option

    Other Parameters
    ----------------
    name : str
        The name of the option. If not provided, the name of the parameter will be used.
    name_localizations : Mapping[hikari.Locale, str]
        The name of the option in different locales
    description_localizations : Mapping[hikari.Locale, str]
        The description of the option in different locales
    min_length : int | None
        The minimum length of the option
    max_length : int | None
        The maximum length of the option
    choices : t.Sequence[str | hikari.CommandChoice] | t.Mapping[str, str] | None
        The choices for the option. If provided, these will be the only valid values for the option.
    autocomplete_with : AutocompleteCallbackT[ClientT, str] | None
        The callback that is invoked when the user autocompletes the option

    !!! warning
        You cannot provide both `choices` and `autocomplete_with` at the same time.
    """

    __slots__: t.Sequence[str] = ("_min_length", "_max_length")

    def __init__(
        self,
        description: str = "No description provided.",
        *,
        name: str | None = None,
        name_localizations: t.Mapping[hikari.Locale, str] | None = None,
        description_localizations: t.Mapping[hikari.Locale, str] | None = None,
        min_length: int | None = None,
        max_length: int | None = None,
        choices: t.Sequence[str | hikari.CommandChoice] | t.Mapping[str, str] | None = None,
        autocomplete_with: AutocompleteCallbackT[ClientT, str] | None = None,
    ) -> None:
        super().__init__(
            name=name,
            description=description,
            name_localizations=name_localizations,
            description_localizations=description_localizations,
            choices=choices,
            autocomplete_with=autocomplete_with,
        )
        self._min_length = min_length
        self._max_length = max_length

    @property
    def min_length(self) -> int | None:
        """The minimum length of the option."""
        return self._min_length

    @property
    def max_length(self) -> int | None:
        """The maximum length of the option."""
        return self._max_length


@attr.define(slots=True, kw_only=True)
class StrOption(OptionWithChoices[str, ClientT, StrParams[ClientT]]):
    """A slash command option that represents a string.

    ??? hint
        To add an option of this type to your command, add an argument to your command function with the following type hint:
        ```py
        opt_name: arc.Option[str, StrParams(...)]
        ```
    """

    min_length: int | None = None
    """The minimum length of the option."""
    max_length: int | None = None
    """The maximum length of the option."""

    @property
    def option_type(self) -> OptionType:
        return OptionType.STRING

    @classmethod
    def _from_params(cls, *, name: str, is_required: bool, params: StrParams[ClientT], **kwargs: t.Any) -> te.Self:
        return cls(
            name=name,
            description=params.description,
            is_required=is_required,
            min_length=params.min_length,
            max_length=params.max_length,
            choices=params.choices,
            autocomplete_with=params.autocomplete_with,
            name_localizations=params.name_localizations,
            description_localizations=params.description_localizations,
            **kwargs,
        )

    def _to_dict(self) -> dict[str, t.Any]:
        return {**super()._to_dict(), "min_length": self.min_length, "max_length": self.max_length}


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from alluka.abc import Client as Injector

from .client import Client
from .command import CallableCommandBase, CallableCommandProto, CommandBase, CommandProto
from .concurrency_limiting import ConcurrencyLimiterProto, HasConcurrencyLimiter, with_concurrency_limit
from .error_handler import HasErrorHandler
from .hookable import Hookable, HookResult, with_hook, with_post_hook
from .limiter import LimiterProto
from .option import (
    CommandOptionBase,
    Option,
    OptionBase,
    OptionParams,
    OptionType,
    OptionWithChoices,
    OptionWithChoicesParams,
)
from .plugin import PluginBase

__all__ = (
    "Injector",
    "HasErrorHandler",
    "HasConcurrencyLimiter",
    "CommandBase",
    "CommandProto",
    "CallableCommandProto",
    "CallableCommandBase",
    "Option",
    "OptionBase",
    "OptionType",
    "CommandOptionBase",
    "OptionParams",
    "OptionWithChoices",
    "OptionWithChoicesParams",
    "Client",
    "PluginBase",
    "Hookable",
    "HookResult",
    "LimiterProto",
    "ConcurrencyLimiterProto",
    "with_hook",
    "with_post_hook",
    "with_concurrency_limit",
)

# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import hikari

import arc

bot = hikari.GatewayBot("...", banner=None)
client = arc.GatewayClient(
    bot, autodefer=arc.AutodeferMode.EPHEMERAL, default_permissions=hikari.Permissions.MANAGE_GUILD, is_dm_enabled=False
)

plugin = arc.GatewayPlugin("foo", default_permissions=hikari.Permissions.NONE, is_dm_enabled=True)


@plugin.include
@arc.slash_command("foo", is_dm_enabled=False)
async def foo(ctx: arc.GatewayContext) -> None:
    await ctx.respond("foo")


@plugin.include
@arc.user_command("bar")
async def bar(ctx: arc.GatewayContext, user: hikari.User) -> None:
    await ctx.respond(f"bar {user.mention}")


group = plugin.include_slash_group("group", "group description", autodefer=False)


@group.include
@arc.slash_subcommand("qux")
async def qux(ctx: arc.GatewayContext) -> None:
    await ctx.respond("foo")


client.add_plugin(plugin)


@client.include
@arc.message_command("baz", autodefer=True)
async def baz(ctx: arc.GatewayContext, message: hikari.Message) -> None:
    await ctx.respond(f"baz {message.author.mention}")


def test_settings_inheritance() -> None:
    assert foo.is_dm_enabled is False
    assert foo.default_permissions == hikari.Permissions.NONE
    assert foo.autodefer is arc.AutodeferMode.EPHEMERAL

    assert bar.is_dm_enabled is True
    assert bar.default_permissions == hikari.Permissions.NONE
    assert bar.autodefer is arc.AutodeferMode.EPHEMERAL

    assert baz.is_dm_enabled is False
    assert baz.default_permissions == hikari.Permissions.MANAGE_GUILD
    assert baz.autodefer is arc.AutodeferMode.ON

    assert qux.autodefer is arc.AutodeferMode.OFF



import os
import re
import types
import typing as t

from setuptools import find_namespace_packages, setup

name = "arc"


def parse_meta() -> types.SimpleNamespace:
    with open(os.path.join(name, "internal", "about.py")) as fp:
        code = fp.read()

    token_pattern = re.compile(
        r"^__(?P<key>\w+)?__\s*:?.*=\s*(?P<quote>(?:'{3}|\"{3}|'|\"))(?P<value>.*?)(?P=quote)", re.M
    )

    groups = {}

    for match in token_pattern.finditer(code):
        group = match.groupdict()
        groups[group["key"]] = group["value"]

    return types.SimpleNamespace(**groups)


def long_description() -> str:
    with open("README.md") as fp:
        return fp.read()


def parse_requirements_file(path: str) -> t.List[str]:
    with open(path) as fp:
        dependencies = (d.strip() for d in fp.read().split("\n") if d.strip())
        return [d for d in dependencies if not d.startswith("#")]


meta = parse_meta()

setup(
    name="hikari-arc",
    version=meta.version,
    description="A command handler for hikari with a focus on type-safety and correctness.",
    long_description=long_description(),
    long_description_content_type="text/markdown",
    author=meta.author,
    author_email=meta.author_email,
    maintainer=meta.maintainer,
    url=meta.url,
    packages=find_namespace_packages(include=[name + "*"]),
    package_data={"arc": ["py.typed"]},
    license=meta.license,
    include_package_data=True,
    zip_safe=False,
    install_requires=parse_requirements_file("requirements.txt"),
    extras_require={
        ':sys_platform=="win32"': ["colorama"],
        "docs": parse_requirements_file("doc_requirements.txt"),
        "dev": parse_requirements_file("dev_requirements.txt"),
        "rest": parse_requirements_file("rest_requirements.txt"),
        "cron": parse_requirements_file("cron_requirements.txt"),
    },
    python_requires=">=3.10.0,<3.13",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)

# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import abc
import typing as t

from arc.internal.types import ClientT, ErrorHandlerCallbackT

if t.TYPE_CHECKING:
    from ..context import Context


class HasErrorHandler(abc.ABC, t.Generic[ClientT]):
    """A trait for objects that can have an error handler set on them."""

    __slots__: t.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def error_handler(self) -> ErrorHandlerCallbackT[ClientT] | None:
        """The error handler for this object."""

    @error_handler.setter
    @abc.abstractmethod
    def error_handler(self, callback: ErrorHandlerCallbackT[ClientT] | None) -> None:
        """Set the error handler for this object."""

    @t.overload
    def set_error_handler(
        self, callback: None = ...
    ) -> t.Callable[[ErrorHandlerCallbackT[ClientT]], ErrorHandlerCallbackT[ClientT]]:
        ...

    @t.overload
    def set_error_handler(self, callback: ErrorHandlerCallbackT[ClientT]) -> ErrorHandlerCallbackT[ClientT]:
        ...

    def set_error_handler(
        self, callback: ErrorHandlerCallbackT[ClientT] | None = None
    ) -> ErrorHandlerCallbackT[ClientT] | t.Callable[[ErrorHandlerCallbackT[ClientT]], ErrorHandlerCallbackT[ClientT]]:
        """Decorator to set an error handler for this object. This can be added to commands, groups, or plugins.

        This function will be called when an exception is raised during the invocation of a command.

        Examples
        --------
        ```py
        @client.include
        @arc.slash_command("foo", "Foo command description")
        async def foo(ctx: arc.GatewayContext) -> None:
            raise RuntimeError("foo")

        @foo.set_error_handler
        async def foo_error_handler(ctx: arc.GatewayContext, exc: Exception) -> None:
            if isinstance(exc, RuntimeError):
                await ctx.respond("foo failed")
                return

            raise exc
        ```

        !!! warning
            Errors that cannot be handled by the error handler should be re-raised.
            Otherwise they will not propagate to the next error handler.
        """

        def decorator(func: ErrorHandlerCallbackT[ClientT]) -> ErrorHandlerCallbackT[ClientT]:
            self.error_handler = func
            return func

        if callback is not None:
            return decorator(callback)

        return decorator

    @abc.abstractmethod
    async def _handle_exception(self, ctx: Context[ClientT], exc: Exception) -> None:
        """Handle an exception or propagate it to the next error handler if it cannot be handled."""


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import typing as t

from arc.abc.limiter import LimiterProto
from arc.context.base import Context
from arc.errors import UnderCooldownError
from arc.internal.types import ClientT
from arc.utils.ratelimiter import RateLimiter, RateLimiterExhaustedError

__all__ = (
    "LimiterHook",
    "global_limiter",
    "guild_limiter",
    "channel_limiter",
    "user_limiter",
    "member_limiter",
    "custom_limiter",
)


class LimiterHook(RateLimiter[Context[ClientT]], LimiterProto[ClientT]):
    """The default implementation of a ratelimiter that can be used as a hook.

    Parameters
    ----------
    period : float
        The period, in seconds, after which the bucket resets.
    limit : int
        The amount of requests allowed in a bucket.
    get_key_with : t.Callable[[Context[t.Any]], str]
        A callable that returns a key for the ratelimiter bucket.

    See Also
    --------
    - [`global_limiter`][arc.utils.hooks.limiters.global_limiter]
    - [`guild_limiter`][arc.utils.hooks.limiters.guild_limiter]
    - [`channel_limiter`][arc.utils.hooks.limiters.channel_limiter]
    - [`user_limiter`][arc.utils.hooks.limiters.user_limiter]
    - [`member_limiter`][arc.utils.hooks.limiters.member_limiter]
    - [`custom_limiter`][arc.utils.hooks.limiters.custom_limiter]
    """

    async def acquire(self, ctx: Context[ClientT], /, *, wait: bool = True) -> None:
        """Acquire a bucket, block execution if ratelimited and wait is True.

        Parameters
        ----------
        ctx : Context[t.Any]
            The context to evaluate the ratelimit under.
        wait : bool
            Determines if this call should block in
            case of hitting a ratelimit.

        Raises
        ------
        UnderCooldownError
            If the ratelimiter is ratelimited and wait is False.
        """
        try:
            return await super().acquire(ctx, wait=wait)
        except RateLimiterExhaustedError as exc:
            raise UnderCooldownError(
                self, exc.retry_after, f"Command is under cooldown for '{exc.retry_after}' seconds."
            ) from exc


def global_limiter(period: float, limit: int) -> LimiterHook[t.Any]:
    """Create a global ratelimiter.

    This ratelimiter is shared across all invocation contexts.

    Parameters
    ----------
    period : float
        The period, in seconds, after which the bucket resets.
    limit : int
        The amount of requests allowed in a bucket.

    Examples
    --------
    ```py
    @arc.with_hook(arc.global_limiter(5.0, 1))
    ```
    """
    return LimiterHook(period, limit, get_key_with=lambda _: "amongus")


def guild_limiter(period: float, limit: int) -> LimiterHook[t.Any]:
    """Create a guild ratelimiter.

    This ratelimiter is shared across all contexts in a guild.

    Parameters
    ----------
    period : float
        The period, in seconds, after which the bucket resets.
    limit : int
        The amount of requests allowed in a bucket.

    Examples
    --------
    ```py
    @arc.with_hook(arc.guild_limiter(5.0, 1))
    ```
    """
    return LimiterHook(period, limit, get_key_with=lambda ctx: str(ctx.guild_id))


def channel_limiter(period: float, limit: int) -> LimiterHook[t.Any]:
    """Create a channel ratelimiter.

    This ratelimiter is shared across all contexts in a channel.

    Parameters
    ----------
    period : float
        The period, in seconds, after which the bucket resets.
    limit : int
        The amount of requests allowed in a bucket.

    Examples
    --------
    ```py
    @arc.with_hook(arc.channel_limiter(5.0, 1))
    ```
    """
    return LimiterHook(period, limit, get_key_with=lambda ctx: str(ctx.channel_id))


def user_limiter(period: float, limit: int) -> LimiterHook[t.Any]:
    """Create a user ratelimiter.

    This ratelimiter is shared across all contexts by a user.

    Parameters
    ----------
    period : float
        The period, in seconds, after which the bucket resets.
    limit : int
        The amount of requests allowed in a bucket.

    Examples
    --------
    ```py
    @arc.with_hook(arc.user_limiter(5.0, 1))
    ```
    """
    return LimiterHook(period, limit, get_key_with=lambda ctx: str(ctx.author.id))


def member_limiter(period: float, limit: int) -> LimiterHook[t.Any]:
    """Create a member ratelimiter.

    This ratelimiter is shared across all contexts by a member in a guild.
    The same user in a different guild will be assigned a different bucket.

    Parameters
    ----------
    period : float
        The period, in seconds, after which the bucket resets.
    limit : int
        The amount of requests allowed in a bucket.

    Examples
    --------
    ```py
    @arc.with_hook(arc.member_limiter(5.0, 1))
    ```
    """
    return LimiterHook(period, limit, get_key_with=lambda ctx: f"{ctx.author.id}:{ctx.guild_id}")


def custom_limiter(period: float, limit: int, get_key_with: t.Callable[[Context[t.Any]], str]) -> LimiterHook[t.Any]:
    """Create a ratelimiter with a custom key extraction function.

    Parameters
    ----------
    period : float
        The period, in seconds, after which the bucket resets.
    limit : int
        The amount of requests allowed in a bucket.
    get_key_with : t.Callable[[Context[t.Any]], str]
        A callable that returns a key for the ratelimiter bucket. This key is used to identify the bucket.
        For instance, to create a ratelimiter that is shared across all contexts in a guild,
        you would use `lambda ctx: str(ctx.guild_id)`.

    Examples
    --------
    ```py
    # This is identical to 'arc.guild_limiter(5.0, 1)'
    @arc.with_hook(arc.custom_limiter(5.0, 1, lambda ctx: str(ctx.guild_id)))
    ```
    """
    return LimiterHook(period, limit, get_key_with=get_key_with)


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import typing as t

import attr

from arc.abc.option import OptionType, OptionWithChoices, OptionWithChoicesParams
from arc.internal.types import ClientT

if t.TYPE_CHECKING:
    import hikari
    import typing_extensions as te

    from ...internal.types import AutocompleteCallbackT

__all__ = ("FloatOption", "FloatParams")


@t.final
class FloatParams(OptionWithChoicesParams[float, ClientT]):
    """The parameters for a float option.

    Parameters
    ----------
    description : str
        The description of the option

    Other Parameters
    ----------------
    name : str
        The name of the option. If not provided, the name of the parameter will be used.
    name_localizations : Mapping[hikari.Locale, str] | None
        The name of the option in different locales
    description_localizations : Mapping[hikari.Locale, str] | None
        The description of the option in different locales
    min : float | None
        The minimum value of the option
    max : float | None
        The maximum value of the option
    choices : t.Sequence[float | hikari.CommandChoice] | t.Mapping[str, float] | None
        The choices for the option. If provided, these will be the only valid values for the option.
    autocomplete_with : AutocompleteCallbackT[ClientT, int] | None
        The callback that is invoked when the user autocompletes the option

    !!! warning
        You cannot provide both `choices` and `autocomplete_with` at the same time.
    """

    __slots__: t.Sequence[str] = ("_min", "_max")

    def __init__(
        self,
        description: str = "No description provided.",
        *,
        name: str | None = None,
        name_localizations: t.Mapping[hikari.Locale, str] = {},
        description_localizations: t.Mapping[hikari.Locale, str] = {},
        min: float | None = None,
        max: float | None = None,
        choices: t.Sequence[float | hikari.CommandChoice] | t.Mapping[str, float] | None = None,
        autocomplete_with: AutocompleteCallbackT[ClientT, float] | None = None,
    ) -> None:
        super().__init__(
            name=name,
            description=description,
            name_localizations=name_localizations,
            description_localizations=description_localizations,
            choices=choices,
            autocomplete_with=autocomplete_with,
        )
        self._min = min
        self._max = max

    @property
    def min(self) -> float | None:
        """The minimum value of the option."""
        return self._min

    @property
    def max(self) -> float | None:
        """The maximum value of the option."""
        return self._max


@attr.define(slots=True, kw_only=True)
class FloatOption(OptionWithChoices[float, ClientT, FloatParams[ClientT]]):
    """A slash command option that represents a float.

    ??? hint
        To add an option of this type to your command, add an argument to your command function with the following type hint:
        ```py
        opt_name: arc.Option[float, FloatParams(...)]
        ```
    """

    min: float | None = None
    """The minimum value of the option."""
    max: float | None = None
    """The maximum value of the option."""

    @property
    def option_type(self) -> OptionType:
        return OptionType.FLOAT

    @classmethod
    def _from_params(cls, *, name: str, is_required: bool, params: FloatParams[ClientT], **kwargs: t.Any) -> te.Self:
        return cls(
            name=name,
            description=params.description,
            is_required=is_required,
            min=params.min,
            max=params.max,
            choices=params.choices,
            autocomplete_with=params.autocomplete_with,
            name_localizations=params.name_localizations,
            description_localizations=params.description_localizations,
            **kwargs,
        )

    def _to_dict(self) -> dict[str, t.Any]:
        return {**super()._to_dict(), "min_value": self.min, "max_value": self.max}


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import typing as t

import hikari

from arc.abc.option import OptionType

if t.TYPE_CHECKING:
    from arc.abc.option import CommandOptionBase
    from arc.internal.types import ClientT


OPTIONTYPE_TO_TYPE: dict[OptionType, type[t.Any]] = {
    OptionType.STRING: str,
    OptionType.INTEGER: int,
    OptionType.BOOLEAN: bool,
    OptionType.USER: hikari.PartialUser,
    OptionType.CHANNEL: hikari.PartialChannel,
    OptionType.ROLE: hikari.Role,
    OptionType.MENTIONABLE: hikari.Unique,
    OptionType.FLOAT: float,
    OptionType.ATTACHMENT: hikari.Attachment,
}
"""Used for runtime type checking in Context.get_option, not much else at the moment."""


def resolve_snowflake_value(
    value: hikari.Snowflake, opt_type: hikari.OptionType | int, resolved: hikari.ResolvedOptionData
) -> t.Any:
    """Resolve a snowflake value into a resolved option.

    Parameters
    ----------
    value : hikari.Snowflake
        The snowflake value to resolve.
    opt_type : hikari.OptionType | int
        The type of the option.
    resolved : hikari.ResolvedOptionData
        The resolved option data of the interaction.

    Returns
    -------
    Any
        The resolved snowflake value.

    Raises
    ------
    ValueError
        If the option type is not a valid option type.
    """
    match opt_type:
        case hikari.OptionType.USER:
            out = resolved.members.get(value) or resolved.users[value]
        case hikari.OptionType.ATTACHMENT:
            out = resolved.attachments[value]
        case hikari.OptionType.CHANNEL:
            out = resolved.channels[value]
        case hikari.OptionType.ROLE:
            out = resolved.roles[value]
        case hikari.OptionType.MENTIONABLE:
            out = resolved.members.get(value) or resolved.users.get(value) or resolved.roles[value]
        case _:
            raise ValueError(f"Unexpected option type '{opt_type}.'")

    return out


def resolve_options(
    local_options: t.MutableMapping[str, CommandOptionBase[ClientT, t.Any, t.Any]],
    incoming_options: t.Sequence[hikari.CommandInteractionOption],
    resolved: hikari.ResolvedOptionData | None,
) -> dict[str, t.Any]:
    """Resolve the options into kwargs for the callback.

    Parameters
    ----------
    local_options : t.MutableMapping[str, Option[t.Any, t.Any]]
        The options of the locally stored command.
    incoming_options : t.Sequence[hikari.CommandInteractionOption]
        The options of the interaction.
    resolved : hikari.ResolvedOptionData
        The resolved option data of the interaction.

    Returns
    -------
    dict[str, Any]
        The resolved options as kwargs, ready to be passed to the callback.
    """
    option_kwargs: dict[str, t.Any] = {}

    for arg_name, opt in local_options.items():
        inter_opt = next((o for o in incoming_options if o.name == opt.name), None)

        if inter_opt is None:
            continue

        if isinstance(inter_opt.value, hikari.Snowflake) and resolved:
            option_kwargs[arg_name] = resolve_snowflake_value(inter_opt.value, inter_opt.type, resolved)

        elif isinstance(inter_opt.value, hikari.Snowflake):
            raise ValueError(f"Missing resolved option data for '{inter_opt.name}'.")
        else:
            option_kwargs[arg_name] = inter_opt.value

    return option_kwargs
